#!/usr/bin/perl -w

# Program info2parser.pl

# (c) Gary C. Kessler, 2012 [gck@garykessler.net]

# Program to parse an NTFS INFO2 Record

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

# Initialize global variables

$version = "1.4";
$build_date = "9 April 2013";

# *** This program is an iterative, work-in-progress. ***

# Initialize variables...

# $source is the name of the file to read. It is expected to be
# a set of one or more 1024-byte FILE records

$source = "";
$verbose = 0;

#
# Begin program activity!!
#

print "\nINFO2 Parser V$version - Gary C. Kessler ($build_date)\n\n";

if (parsecommandline ())
  { exit 1; };

$filesize = -s ($source);
print "Source file $source ($filesize bytes)\n\n";

open (INFILE,"<",$source);
binmode (INFILE);

# Read the file header, then read one entry at a time....

print "==================== HEADER ====================\n";

for ($i=0; $i<16; $i++)
  {
  read (INFILE,$byte,1);
  $info2 [$i] = unpack ("C", $byte);
  }

if ($verbose == 1)
  { 
  print "RAW DATA\n";
  list_bytes (0,16,@info2);
  }

# Bytes 0-3 are the file header (0x00000005)

  $bpos = 0; $len = 4;
  $temp = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  File header: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  if ($temp == 5)
      { print " [valid]\n"; }
    else
      { print " [invalid]\n"; }

# Bytes 4-7 are the number of entries in INFO2 file

  $bpos += $len; $len = 4;
  $num_entries = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  Number of entries in file: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print " [", commify ($num_entries), "]\n";

# Bytes 8-11 are undocumented

  $bpos += $len; $len = 4;
  $temp = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  (Undocumented): ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print " [", commify ($temp), "]\n";

# Bytes 12-15 are the size of each INFO2 File Entry (bytes); usually 0x320 (800)

  $bpos += $len; $len = 4;
  $entry_size = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  Size of INFO2 file entry: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print " [", commify ($entry_size), " bytes]\n\n";

# If entry size is not 800, then dump the rest of the file and exit

if ($entry_size != 800)
  {
  print "\nNon-standard INFO2 entry size; cannot parse\n";
  exit;
  }

# Now, read each entry and parse

for ($i=1; $i<=$num_entries; $i++)
  {
  for ($j=0; $j<800; $j++)
    {
    read (INFILE,$byte,1);
    $info2 [$j] = unpack ("C", $byte);
    }

  print "==================== INFO2 Entry #", $i-1, " ====================\n";

  if ($verbose == 1)
    { 
    print "RAW DATA\n";
    list_bytes (0,800,@info2);
    }

# Bytes 0-4 are not documented!!!! (The format of the record is incorrect in Carrier.)

  $bpos = 0; $len = 4;
  $temp = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  (Undocumented): ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print "\n";

# Bytes 5-264 are the full path name of the deleted file (ASCII)

  $bpos += $len; $len = 260;
  printf ("%03d-%03d  Full path name of deleted file (ASCII):\n", $bpos, $bpos+$len-1);
  list_bytes ($bpos,$len,@info2);

  for ($k=0; $k<$len; $k++)
    {
    if ($info2 [$bpos+$k] > 0)
        { print chr ($info2 [$bpos+$k]); }
      else
        { $k = $len; }
    }

  print "\n\n";

# Bytes 264-267 are the unique index number of this entry

  $bpos += $len; $len = 4;
  $temp = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  Unique index number: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print " [", commify ($temp), "]\n";

# Bytes 268-271 are the drive number from where file was deleted (0=A, 1=B, 2=C, etc.)

  $bpos += $len; $len = 4;
  $temp = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  Drive number from where file was deleted: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print " [", chr ($temp+0x41), "]\n";

# Bytes 272-279 are the date/time when file was deleted

  $bpos += $len; $len = 8;
  $temp = little_endian ($bpos,$len,@info2);
  $temp = $temp / 10000000 - 11644473600;
  printf ("%03d-%03d  Date/time when file was deleted: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print "\n         ", scalar (gmtime ($temp)),"\n";

# Bytes 280-283 are the size of the deleted file (bytes)

  $bpos += $len; $len = 4;
  $temp = little_endian ($bpos,$len,@info2);
  printf ("%03d-%03d  Size of deleted file: ", $bpos, $bpos+$len-1);
  list_hex_digits ($bpos,$len,@info2);
  print " [", commify ($temp), " bytes]\n";

# 284-799, Full path name of deleted file (Unicode)

  $bpos += $len; $len = 516;
  printf ("%03d-%03d  Full path name of deleted file (Unicode):\n", $bpos, $bpos+$len-1);
  list_bytes ($bpos,$len,@info2);

  for ($k=0; $k<$len; $k++)
    {
    if ($info2 [$bpos+$k*2] > 0)
        { print chr (little_endian ($bpos+$k*2,2,@info2)); }
      else
        { $k = $len; }
    }

  print "\n\n";
  }

close (INFILE);



# ********************************************************
# commify
# ********************************************************

# Routine from PerlFAQ5 -- add commas to large integers
# http://perldoc.perl.org/perlfaq5.html
  #How-can-I-output-my-numbers-with-commas-added?

sub commify
{
local $_  = shift;
1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
return $_;
}


# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: info2parser [-i input_file] [-v]
               info2parser {[-h] | [-t]}

 where: -i is the input file name
        -h prints this help file
        -t displays the template file
        -v (verbose) displays the entire INFO2 record in hex

If no -i switch is provided, the program will ask for the
input file name.

The input file should be an INFO2 file in binary form.

EOT
return;
}

# ********************************************************
# list_ASCII
# ********************************************************

# Given an array name, starting position, and number of digits, print
# out an ASCII character string

sub list_ASCII ($$@)
{
my ($start,$n,@array) = @_;
my $i;

for ($i = $start; $i < $start+$n; $i++)
  {
  print chr ($array[$i]);
  }

return;
}

# ********************************************************
# list_bytes
# ********************************************************

# Given the array and a number of bytes, set the output to 
# dump X 16-byte lines followed by a single line with the
# rest of the bytes

sub list_bytes
{
my ($start,$n,@array) = @_;
my ($j, $k, $full_lines, $short_line);

# Determine how many full lines to print, assuming 16 B/line

$full_lines = int ($n/16);
$short_line = $n - ($full_lines * 16);

$k = $start;

for ($j=0; $j<$full_lines; $j++)
  {
  set_output ($k,16,@array);
  $k += 16;
  }

if ($short_line > 0)
  { set_output ($k,$short_line,@array); }

print "\n";
}

# ********************************************************
# list_hex_digits
# ********************************************************

# Given an array name, starting position, and number of digits, print
# out a string of hex digits in the format 0xAA-BB-CC...

sub list_hex_digits
{
my ($start,$n,@array) = @_;
my $i;

for ($i = $start; $i < $start+$n; $i++)
  {
  if ($i == $start)
      { print "0x"; }
    else
      { print "-"; }
  printf ("%02X", $array[$i]);
  }

return;
}

# ********************************************************
# little_endian
# ********************************************************

# Given an array name, starting position, and number of digits,
# calculate the decimal value using little endian (LSB first)

sub little_endian
{
my ($start,$n,@array) = @_;
my ($i, $sum);

$sum = $array[$start];
for ($i=1; $i<$n; $i++)
  { $sum += $array[$start+$i] * (2**($i*8)); }

return $sum;
}

# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present.

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)
  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      $ARGV[$i] eq "-v" && do
         {
         $verbose = 1;
         last PARMS;
         };
      $ARGV[$i] eq "-t" && do
         {
         $layout = "layout_INFO2.txt";
         open (INFILE, "<", $layout)
          or die "Can't open input file \"$layout\": $!";
         while (not (eof INFILE))
           {
           $line = <INFILE>;
           print $line;
           }
         close INFILE;
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: info2parser [-i input_file] [-v]\n";
         print "       info2parser {[-h] | [-t]}\n\n";
         $state = 1;
         return $state;
         };
      };
    };
  };

# Prompt for file names, if not supplied

if ($source eq "")
  {
  print "Enter the input file name: ";
  chomp ($source = <STDIN>);
  print "\n";
  };

return $state;
}

# ********************************************************
# set_output
# ********************************************************

# Print output of the data bytes to show both hex and ASCII
# values.

sub set_output ($$@)
{
my ($start,$n,@array) = @_;
my ($j, $k, $linemax);

# Format for a maximum of 16 items per line

$linemax = 16;

printf ("%03d: ", $start);
for ($j=0; $j<$n; $j++)
  {
  printf (" %02X", $array[$start+$j]);
  }
if ($n < $linemax)
  {
  for ($j=1; $j<=$linemax-$n; $j++)
    { print "   "; };
  }

print "   ";
for ($j=0; $j<$n; $j++)
  {
  $k = $array[$start+$j];

#   Print a dot (.) for unprintable characters

  if ($k <= 0x1f || ($k >= 0x7f && $k <= 0x9f))
      { print "."; }
    else
      { print chr($k); }
  }
print "\n";

return;
}

